{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Statically Determinate 2D Trusses\n",
    "This notebook developes a function that will compute the forces and reactions in a \n",
    "statically determinate 2D truss and display the results. Inputs are:\n",
    "* a list of joint names (ids) and x, y coordinates.\n",
    "* a list of member incidences - the pairs of joints that are connected by truss members.\n",
    "* a description of the reaction components, giving joint id and direction for each.\n",
    "* a set of loads expressed as forces on joints, ghiving joint id, magnitude and direction for each."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from numpy.linalg import solve, matrix_rank\n",
    "from math import sqrt, atan2, degrees"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "class SDTError(Exception):\n",
    "    pass"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Joints\n",
    "Each joint object remebers its id and x and y coordinates.  A number from 0 to number of joints-1\n",
    "is sequentially assigned as each joint is created.  Each joint is placed in the 'Joints' dictionary,\n",
    "indexed by joint id.  We check to ensure that all ids are unique, and that no two\n",
    "joints have the same coordinates."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "Joints = dict()   # global dictionary of Joints\n",
    "class Joint:\n",
    "    \n",
    "    def __init__(self,jid,x,y):\n",
    "        if jid in Joints:\n",
    "            raise SDTError(\"Joint '{}' is defined more than once.\".format(jid))\n",
    "        for oj in Joints.values():\n",
    "            if oj.x == x and oj.y == y:\n",
    "                raise SDTError(\"Joint '{}' has the same coordinates as '{}'.\".format(jid,oj.jid))\n",
    "        self.jid = jid\n",
    "        self.n = len(Joints)   # assign sequential number\n",
    "        self.x = x\n",
    "        self.y = y\n",
    "        Joints[self.jid] = self   # add new joint to the Joints dictionary"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Members\n",
    "Each truss member is defined by simply naming the joints at either end of the member.  An internal id\n",
    "is generated by joing the two joint ids together (i.e., a member from joint _d_ to joint _g_ would have the\n",
    "member id _dg_.  An number from 0 to n-1 is sequentially assigned to each new member in the\n",
    "order they are encountered.  Member objects are placed in the global 'Unknowns' dictionary, indexed by member id.\n",
    "\n",
    "Each member has attributes:\n",
    "* _uid_ - the member id\n",
    "* _n_ - member (or unknown) number, starting at 0\n",
    "* _ji_ - the Joint object at the _i_ end (the first end of the pair of joints)\n",
    "* _jj_ - the Joint object at the _j_ end\n",
    "* _L_ - the length of the member\n",
    "* _l_ - a directional cosine - the cosine of the angle between the member axis and the x-axis.\n",
    "* _m_ - a directional cosine - the cosine of the angle between the member axis and the y-axis.\n",
    "\n",
    "The following errors are checked as each new Member object is created:\n",
    "1. the joints on either end must be different\n",
    "1. the joint ids have been previously input (Joint objects exist for both ends)\n",
    "1. the member has not been mentioned more than once\n",
    "1. the length of the member is not 0."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [],
   "source": [
    "Unknowns = dict()     # dictionary of objects representing unknowns (member forces and reactions)\n",
    "JointPairs = dict()   # dictionary of pairs of joint ids for asll members - indexed by a tuple of joint ids (jid1,jid2)  jid1 <= jid2\n",
    "\n",
    "class Member:\n",
    "    \n",
    "    def __init__(self,jid_i,jid_j):\n",
    "        mid = str(jid_i) + str(jid_j)\n",
    "        if jid_i == jid_j:\n",
    "            raise SDTError(\"Member '{}' has both joints the same: '{}'.\".format(mid,jid_i))\n",
    "        for jid in [jid_i,jid_j]:\n",
    "            if jid not in Joints:\n",
    "                raise SDTError(\"Joint '{}' for member '{}' is not defined.\".format(jid,mid))\n",
    "        if mid in Unknowns:\n",
    "            raise SDTError(\"Member '{}' is defined more than once.\".format(mid))\n",
    "        # remember each pair of connected joints, but make sure the names are in alphabetic order\n",
    "        jpair = (jid_i,jid_j) if jid_i <= jid_j else (jid_j,jid_i)\n",
    "        if jpair in JointPairs:\n",
    "            raise SDTError(\"Joints '{}' and '{}' are also connected by member '{}'.\".format(jid_i,jid_j,JointPairs[jpair].uid))\n",
    "            \n",
    "        ji = Joints[jid_i]   # retrieve joint objects for both ends\n",
    "        jj = Joints[jid_j]\n",
    "        dx = jj.x - ji.x     # coordinate differences\n",
    "        dy = jj.y - ji.y\n",
    "        L = sqrt(dx*dx + dy*dy)\n",
    "        if L == 0.:\n",
    "            raise SDTError(\"Length of member '{}' is zero.\".format(mid))\n",
    "            \n",
    "        l = dx/L             # direction cosines\n",
    "        m = dy/L\n",
    "        \n",
    "        self.uid = mid            # store id of member (or 'unknown')\n",
    "        self.n = len(Unknowns)    # assign sequential number\n",
    "        self.ji = ji              # remeber end joint objects\n",
    "        self.jj = jj\n",
    "        self.L = L                # remember geometry\n",
    "        self.l = l\n",
    "        self.m = m\n",
    "        \n",
    "        Unknowns[self.uid] = self  # store in Unknowns dictionary\n",
    "        JointPairs[jpair] = self   # and also in joint pairs dictionary"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Reactions\n",
    "Reactions are also unknowns.  They are defined by specifying a direction of the reaction force at a joint.\n",
    "The direction is specified by giving relative coordinate differences along the line of action of the\n",
    "reaction.  A direction of 3,4  means the line of action rises 4 units in y for every 3 units in x, and the\n",
    "cosine of the angle between the l.o.a and the x-axis would be 0.6 and the cosine wrt to the y-axis\n",
    "would be 0.8.\n",
    "\n",
    "The following errors are checked:\n",
    "1. the named Joint has been previously defined in the joint coordinates input\n",
    "1. the direction is valid (i.e, not 0,0)\n",
    "1. the reaction is not defined more than once\n",
    "\n",
    "Reaction objects have the following attributes:\n",
    "* _jid_ - the id of the joint\n",
    "* _uid_ - the id of the reaction (or unknown) - automatically generated from the joint id and direction\n",
    "* _n_ - the sequential number of the unknown\n",
    "* _ji_ - the Joint object corresponding to the _jid_\n",
    "* _jj_ - None\n",
    "* _l_ - directional cosine (x-axis) of the line of action\n",
    "* _m_ - directional cosine (y-axis) of the line of action\n",
    "\n",
    "Reaction objects are added to the Unknowns dictionary."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Reaction:\n",
    "    \n",
    "    def __init__(self,jid,dx,dy):\n",
    "        if jid not in Joints:\n",
    "            raise SDTError(\"Joint '{}' is not defined for reaction.\".format(jid))\n",
    "            \n",
    "        L = sqrt(dx*dx + dy*dy)\n",
    "        if L == 0.:\n",
    "            raise SDTError(\"dx,dy invalid for joint reaction @ '{}': '{},{}'.\".format(jid,dx,dy))\n",
    "            \n",
    "        # generate the reaction id from the joint id and direction\n",
    "        rid = 'R' + str(jid)\n",
    "        if dx != 0 and dy == 0:\n",
    "            rid += 'x'\n",
    "        elif dy != 0 and dx == 0:\n",
    "            rid += 'y'\n",
    "        else:\n",
    "            a = degrees(atan2(dy,dx))\n",
    "            if a < 0.:\n",
    "                a += 360.\n",
    "            rid += '{:.0f}'.format(a)\n",
    "        if rid in Unknowns:\n",
    "            raise SDTError(\"Reaction '{}' already defined at joint '{}'.\".format(rid,jid))\n",
    "            \n",
    "        self.jid = jid\n",
    "        self.uid = rid\n",
    "        self.n = len(Unknowns)       # sequential unknown number\n",
    "        self.ji = Joints[self.jid]   # the Joint object\n",
    "        self.jj = None\n",
    "        self.l = dx/L                # directional cosines\n",
    "        self.m = dy/L\n",
    "        Unknowns[self.uid] = self"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Putting it all together\n",
    "Function `sdtruss()` accepts as arguments:\n",
    "* _jc_ - a list of joint coordinates.  Each joint is specified by a 3-tuple of its id (usually a character string) \n",
    "  and its _x-_ and _y-_ coordinates.\n",
    "* _mi_ - a list of member incidences.  Each member is specified by a 2-tuple giving the joint ids of the joints\n",
    "  at either end of the membr.\n",
    "* _rf_ - a list of reaction forces.  Each reaction is a 3-tuple giving the joint id and the directions of the line \n",
    "  of action of the reaction force.\n",
    "* _jl_ - a list of joint loads.  Each load is given as a 4-tuple: the joint id, the magnitude of the load, and the directions\n",
    "  of the line of action of the load (directions given as _x-_ and _y-_ relative coordinate differences).\n",
    "* *print_flag* - optional, default `True`.  Whether or not to display a table of resulting meber and reaction forces."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 6,
   "metadata": {},
   "outputs": [],
   "source": [
    "def sdtruss( jc, mi, rf, jl, print_flag=True ):\n",
    "    \n",
    "    # empty all the global dictionaries\n",
    "    global Joints, Unknowns, JointPairs\n",
    "    Joints = dict()\n",
    "    Unknowns = dict()\n",
    "    JointPairs = dict()\n",
    "    connected = set()    # also the set of joints that have been connected to members\n",
    "    \n",
    "    for jid,x,y in jc:        # create all the joints from the joint corrdinates list\n",
    "        Joint(jid,x,y)\n",
    "        \n",
    "    for jidi,jidj in mi:      # create all the members from the member incidences list\n",
    "        Member(jidi,jidj)\n",
    "        connected.add(jidi)   # and keep track of connected joints\n",
    "        connected.add(jidj)\n",
    "        \n",
    "    for jid in Joints:        # now check to ensure that all joints have been connected\n",
    "        if jid not in connected:\n",
    "            raise SDTError(\"Joint '{}' is not connected to any member.\".format(jid))\n",
    "            \n",
    "    for jid,dcx,dcy in rf:    # create all the reaction unknowns.\n",
    "        Reaction(jid,dcx,dcy)\n",
    "        \n",
    "    N = 2*len(Joints)         # now check to ensure that number of unknowns matches number of equations\n",
    "    if len(Unknowns) < N:\n",
    "        raise SDTError(\"Too few unknowns, truss is unstable: {}+{} < 2*{}\".format(len(mi),len(rf),len(jc)))\n",
    "    elif len(Unknowns) > N:\n",
    "        raise SDTError(\"Too many unknowns, truss is statically indeterminate: {}+{} > 2*{}\".format(len(mi),len(rf),len(jc)))\n",
    "        \n",
    "    # create the C matrix from the direction cosines of the member and reaction forces\n",
    "    C = np.zeros((N,N),np.float64)\n",
    "    for unk in Unknowns.values():\n",
    "        k = unk.n             # the column number is the unknown number\n",
    "        ji = unk.ji           # Joint objects give the row numbers\n",
    "        jj = unk.jj\n",
    "        C[ji.n*2,k] = unk.l   # place the direction cosines\n",
    "        C[ji.n*2+1,k] = unk.m\n",
    "        if jj:                # and for members (not reactions), the dcs at the other end\n",
    "            C[jj.n*2,k] = -unk.l\n",
    "            C[jj.n*2+1,k] = -unk.m\n",
    " \n",
    "    # create the load vector from the joint loads input\n",
    "    P = np.zeros((N,1),np.float64)\n",
    "    for jid,p,dcx,dcy in jl:    # for each load\n",
    "        if jid not in Joints:\n",
    "            raise SDTError(\"Joint '{}' not defined in load input.\".format(jid))\n",
    "        L = sqrt(dcx*dcx + dcy*dcy)\n",
    "        if L == 0.:\n",
    "            raise SDTError(\"Improper dcx,dcy for load on joint '{}': {},{}\".format(jid,dcx,dcy))\n",
    "        l = dcx/L               # direction cosines of the line of action\n",
    "        m = dcy/L\n",
    "        j = Joints[jid]         # the joint provides the equation number (row #)\n",
    "        P[j.n*2,0] = -l*p\n",
    "        P[j.n*2+1,0] = -m*p\n",
    "    \n",
    "    # now check for solvabilty and solve for forces, if possible\n",
    "    if matrix_rank(C) < N:\n",
    "        raise SDTError(\"'C' matrix is rank deficient.  Truss is unstable.\")\n",
    "    Q = solve(C,P)\n",
    "    \n",
    "    # optionally display forces and return them    \n",
    "    if not print_flag:\n",
    "        return Q\n",
    "    \n",
    "    print('unknown  joint-i  joint-j        force')\n",
    "    print('-------  -------  -------        -----')\n",
    "    for unk in Unknowns.values():             # for each unknown\n",
    "        uid = unk.uid\n",
    "        iid = unk.ji.jid\n",
    "        jid = unk.jj.jid if unk.jj else ''    # reactions don't have a second joint id\n",
    "        t = Q[unk.n,0]                        # get the force from the unknown number\n",
    "        if jid:                               # if its a member force, display magnitude and C or T\n",
    "            s = 'T' if t >= 0 else 'C'\n",
    "            print(\"{:7s}  {:7s}  {:7s} {:12.4g} {}\".format(uid,iid,jid,abs(t),s))\n",
    "        else:                                 # else its a reaction so display one joint + direction\n",
    "            print(\"{:7s}  {:7s}  {:7s} {:12.4g} {} @ {:.3g},{:.3g}\".format(uid,iid,jid,t,'',unk.l,unk.m))\n",
    "    return Q"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Example Usage\n",
    "\n",
    "![Figure](2019-hw2-q2.svg)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 8,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "unknown  joint-i  joint-j        force\n",
      "-------  -------  -------        -----\n",
      "ab       a        b                 15 C\n",
      "bc       b        c                 75 C\n",
      "ad       a        d              6.667 C\n",
      "bd       b        d                100 C\n",
      "af       a        f              16.41 T\n",
      "ce       c        e                125 T\n",
      "cf       c        f                100 C\n",
      "de       d        e                 60 T\n",
      "ef       e        f                 15 C\n",
      "Rfx      f                  -2.132e-14  @ 1,0\n",
      "Rfy      f                       93.33  @ 0,1\n",
      "Rdy      d                       86.67  @ 0,1\n"
     ]
    }
   ],
   "source": [
    "if __name__ == '__main__':\n",
    "    \n",
    "    # HW2 Q2 2019\n",
    "    \n",
    "    jc = [ ('a', 0, 4),           # joint coordinates (jid,x,y)\n",
    "           ('b', 3, 4),\n",
    "           ('c', 9, 4),\n",
    "           ('d', 0, 0),\n",
    "           ('e', 6, 0),\n",
    "           ('f', 9, 0),\n",
    "           ]\n",
    "\n",
    "    mi = [ ('a','b'),               # member incidences (i,j) joint ids\n",
    "           ('b','c'),\n",
    "           ('a','d'),\n",
    "           ('b','d'),\n",
    "           ('a','f'),\n",
    "           ('c','e'),\n",
    "           ('c','f'),\n",
    "           ('d','e'),\n",
    "           ('e','f'),\n",
    "           ]\n",
    "\n",
    "    rf = [ ('f',1,0),             # reaction forces (jid,dcx,dcy)   dcx,dcy are direction components\n",
    "           ('f',0,1),\n",
    "           ('d',0,1),\n",
    "           ]\n",
    "\n",
    "    jl = [ ('b',80,0,-1),        # joint loads (jid,P,dcx,dcy),\n",
    "           ('e',100,0,-1),\n",
    "           ]\n",
    "\n",
    "    sdtruss( jc, mi, rf, jl )"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.6.8"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
